#       getargs.bash
#       This file is part of zenpakt

#       Copyright 2011 Anshul Narula aka sunny <sunny_delhi18@yahoo.com>
#
#       Zenpakt is free software; you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation; either version 2 of the License, or
#       (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program; if not, write to the Free Software
#       Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#       MA 02110-1301, USA.

#@DOC getargs
#*=========
#* getargs
#*=========
#*
#*-----------------------------
#* parse positional parameters
#*-----------------------------
#*
#*Synopsis
#*--------
#* **getargs** [ **-C** *callback* ] [ **-V** ] -- *arguments*...
#*
#*Description
#*-----------
#*
#*parse the positional parameters according to values specified in array named as
#*getargs. Also return the values in GETARGS array.
#*
#*  **getargs**
#*  getargs must be a pseudo 4-dimensional array with the following
#*  columns for each row (in that order):
#*
#*  * *short*
#*  * *long*
#*  * *argument*
#*  * *repeat*
#*
#*  *short* is the first column which defines a short option.A short option must be
#*  single character long, since only single characters are used to match short options
#*  Thus making a short option either NULL or more than one character ensures that
#*  it would not match any short option, effectively stating that there is no short option
#*  for this case.
#*
#*  *long* defines a long option, it can be any length, if NULL means there is no long
#*  option.
#*
#*  if *argument* is :, it means that the option requires an argument, anything else means
#*  it doesnot require an argument.
#*
#*  *repeat* represents how many times the argument can be repeated, if 0 means that option
#*  should not be present at all.If negative number then, it means it can be repeated
#*  infinite number of times.
#*
#*  **GETARGS**
#*  This is a non-sparse array which contains argument to an option or null.
#*
#*  **GETARGS\_**
#*  This is an associative array which contains the option found as index and space followed by
#*  index of each occurance of argument as element, so that each argument for multiple occurences
#*  may be found and also so that number of spaces in element corresponds to number of occurences
#*  of the given option.
#*
#*Options
#*-------
#*
#* -c callback  call the callback for each option found.
#* -V           do not set variables
#*
#*Return Status
#*-------------
#*
#** 0 success
#** 1 failure
#@ENDDOC

$$

declare -A GETARGS_

getargs()
{
######################
#functions
######################

  #This function calls the callback if set,or returns 0 because no callback is equal to true
  _getargs_callback()
  {
	if [[ ${internal[callback]} ]] ; then\
	  _getargs_callback() { "${internal[callback]}" "$@" ;}
	else
	  _getargs_callback() { return 0 ;}
	fi
	_getargs_callback "$@"
  }

  #check the repeat value
	#if it is a negative number then , repeat is infinite, no need to remove the entry from getargs_short and getargs_long
	#if it is 0 then , limit is exceeded
	#if it is anything else, reduce the repeat values
  #_getargs_repeat index option argument
  _getargs_repeat()
  {
	#if repeat is zero and we reach here , either it is deliberately set repeat to fail in case
  #such an option is supplied or it is an internal error
  #(because the option exceeded the limit of times it must be specified)
  #any way we just return 1
	((${getargs[$1]} == 0 )) && return 1
	if ((internal[assign])) ; then
	  GETARGS+=( "$2" "$3" )
	  GETARGS_["$2"]=" $((${#GETARGS[@]} - 2))"
	fi
  #reduce the repeat each time, for non negative repeat
  ((${getargs[$1]} > 0 ? getargs[$1]-- : 1))
	}

	#parse one or more concatenated short options
	#_getargs_set_args_short anchor option_without_anchor [next_options]
	_getargs_short()
	{
	local i index
	#iterate over the list, each character is/should be a short option
	#if a short option requires an argument, remaining is the argument
	for ((i=0 ; i < ${#2} ; i++)) ; do

	  if [[ ${2:i:1} ]] ; then
	    #check if the option is in the expected option list
			inarray -i index -m 4:1 "${2:i:1}" "${getargs[@]}" || return 1
		  if [[ ${getargs[index+2]} = : ]] ; then
		    if (( i + 1 == ${#2} )) ; then
			    if [[ $3 ]] ; then
					  _getargs_repeat "$((index+3))" "${2:i:1}" "$3" || return 1
					  #call the call back now
					  _getargs_callback "$1" "${2:i:1}" "$3"  || return 1
					  REPLY=2
					  break
			    else
					  print -d "Error: no argument specified for ${2:i:1}"
					  return 1
				  fi

				#use the remaining string as argument
			  else
				  _getargs_repeat "$((index+3))" "${2:i:1}" "${2:i+1}" || return 1
				  _getargs_callback "$2" "${2:i:1}" "${2:++i}" || return 1
				  REPLY=1
				  break
			  fi

			#if we dont need an option
			else
				_getargs_repeat "$((index+3))" "${2:i:1}" || return 1
			  _getargs_callback "$1" "${2:i:1}" || return 1
				REPLY=1
			fi

		else
			print -d "invalid option character: $3"
			return 1
		fi
  done
	}

  #set the long option
  #_getargs_set_args_long anchor option_without_anchor next_argument
	_getargs_long()
	{
	local index
	inarray -i index -m 4:2 "$2" "${getargs[@]}" || return 1
	if [[ ${getargs[index+1]} = : ]] ; then
      [[ $3 ]] || return 1
      _getargs_repeat "$((index+2))" "$2" "$3" || return 1
      _getargs_callback "$1" "$2" "$3" || return 1
      REPLY=2

	else
		_getargs_repeat "$((index+2))" "$2" || return 1
		_getargs_callback "$1" "$2" || return 1
		REPLY=1
	fi
  }

#################
#local variables
#################
declare -A internal int
internal=(
    [assign]=1        #if 0, do not set the array variables,you may want this if you are using a callback
    [anchor_short]=-  #list of | separated short anchors; default anchor is -
    [anchor_long]=--  #list of | separated long anchors; default anchor is --
    [callback]=       #initial callback is nothing
    )

################
#OPTIONS
################
#<irony> parse our own options using getopts
local opt OPTIND=1
while getopts :c:V opt; do
  case $opt in

    c)
      internal[callback]=$OPTARG
      type "${internal[callback]}" >/dev/null 2>&1 || {
				print -d "$FUNCNAME : $getargs_callback not found"
				return 1
		  }
      ;;

    V)
     internal[assign]=0
     ;;

    \?|:)
      return 1
      ;;

   esac
done
shift $((OPTIND-1))
##################
#END OPTIONS
##################

######################
#CHECK ARRAY STRUCTRES
######################
#check if getargs is not empty and is a 4 dimensional array(in multiples of 4)
((${#getargs[@]} ? (${#getargs[@]}%4 == 0) : 0)) || {
  print -d "$FUNCNAME : getargs not defined"
  return 1
  }

######################
#EXTERNAL VARIABLES
######################
#clear up the variables now
if ((internal[assign])) ; then
  GETARGS_OPT=() GETARGS_OPTARG=() GETARGS=() || return 1
fi
GETARGS_IND= || return 1

#######################
#PARSE ARGUMENTS
#######################
local opt i=$# REPLY || return 1
while (($#)) ; do

  #any of the anchors specifying the start of an option alone signals end of options
  #so does anything not starting with these anchors
  [[ $1 = @(--|${internal[anchor_short]}|${internal[anchor_long]}) ]] && shift && break

  #if the option starts with a short anchor and not with a long anchor
  #we may/should have a short option
  if [[ $1 = @(${internal[anchor_short]})* && $1 != @(${internal[anchor_long]})* ]] ; then

    #remove the longest possible matching short anchor
		opt=${1##@(${internal[anchor_short]})}
		_getargs_short "${1%$opt}" "$opt" "${@:2}" || return 1
    shift "$REPLY"

  #if the option starts with a long option option and not a short one
  elif [[ $1 = @(${internal[anchor_long]})* && $1 != @(${internal[anchor_short]})* ]] ; then
		opt=${1##@(${internal[anchor_long]})}
		_getargs_long "${1%$opt}" "$opt" "${@:2}" || return 1
    shift "$REPLY"

  #if option matches both long and short anchors
  elif [[ $1 = @(${internal[anchor_long]})* && $1 = @(${internal[anchor_short]})* ]] ; then

    #choose the longer anchor one
    opt=${1##@(${internal[anchor_long]})}
    opt[1]=${1%$opt}
    opt=${1##@(${internal[anchor_short]})}
    opt=${1%$opt}

    if ((${#opt[0]} > ${#opt[1]})) ; then
      #we have short options
      _getargs_short "$opt" "${1#$opt}" "${@:2}" || return 1
      shift "$REPLY"
    else
      _getargs_long "${opt[1]}" "${1#${opt[1]}}" "${@:2}" || return 1
      shift "$REPLY"
    fi

  else
    #it is end of options
    break
  fi
done
################
#END PARSING
################

#set the index and return
((GETARGS_IND = $# ?  i - $# + 1 : i + 1))
}

